package sun.rmi.server;

import java.security.AccessController;
import sun.security.action.GetBooleanAction;
import java.rmi.server.RemoteCall;
import java.rmi.server.Operation;
import java.rmi.server.RemoteObject;
import sun.misc.SharedSecrets;
import java.io.ObjectInputStream;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import sun.rmi.transport.Connection;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;
import java.io.IOException;
import java.rmi.MarshalException;
import sun.rmi.transport.StreamRemoteCall;
import java.lang.reflect.Method;
import java.rmi.Remote;
import sun.rmi.transport.LiveRef;
import sun.rmi.runtime.Log;
import java.rmi.server.RemoteRef;

public class UnicastRef implements RemoteRef
{
    public static final Log clientRefLog;
    public static final Log clientCallLog;
    private static final long serialVersionUID = 8258372400816541186L;
    protected LiveRef ref;
    
    public UnicastRef() {
    }
    
    public UnicastRef(LiveRef ref) {
        this.ref = ref;
    }
    
    public LiveRef getLiveRef() {
        return this.ref;
    }
    
    @Override
    public Object invoke(Remote remote, Method obj, Object[] array, long n) throws Exception {
        if (UnicastRef.clientRefLog.isLoggable(Log.VERBOSE)) {
            UnicastRef.clientRefLog.log(Log.VERBOSE, "method: " + obj);
        }
        if (UnicastRef.clientCallLog.isLoggable(Log.VERBOSE)) {
            this.logClientCall(remote, obj);
        }
        Connection connection = this.ref.getChannel().newConnection();
        StreamRemoteCall streamRemoteCall = null;
        boolean b = true;
        int n2 = 0;
        try {
            if (UnicastRef.clientRefLog.isLoggable(Log.VERBOSE)) {
                UnicastRef.clientRefLog.log(Log.VERBOSE, "opnum = " + n);
            }
            streamRemoteCall = new StreamRemoteCall(connection, this.ref.getObjID(), -1, n);
            try {
                ObjectOutput outputStream = streamRemoteCall.getOutputStream();
                this.marshalCustomCallData(outputStream);
                Class[] parameterTypes = obj.getParameterTypes();
                for (int i = 0; i < parameterTypes.length; ++i) {
                    marshalValue(parameterTypes[i], array[i], outputStream);
                }
            }
            catch (IOException ex) {
                UnicastRef.clientRefLog.log(Log.BRIEF, "IOException marshalling arguments: ", ex);
                throw new MarshalException("error marshalling arguments", ex);
            }
            streamRemoteCall.executeCall();
            try {
                Class<Object> returnType = obj.getReturnType();
                if (returnType == Void.TYPE) {
                    __Null _Null = null;
                    try {
                        streamRemoteCall.done();
                    }
                    catch (IOException ex3) {
                        b = false;
                    }
                }
                else {
                    Object unmarshalValue = unmarshalValue(returnType, streamRemoteCall.getInputStream());
                    n2 = 1;
                    UnicastRef.clientRefLog.log(Log.BRIEF, "free connection (reuse = true)");
                    this.ref.getChannel().free(connection, true);
                    try {
                        streamRemoteCall.done();
                    }
                    catch (IOException ex2) {
                        b = false;
                    }
                }
            }
            catch (IOException | ClassNotFoundException ex4) {
                ((StreamRemoteCall)streamRemoteCall).discardPendingRefs();
                UnicastRef.clientRefLog.log(Log.BRIEF, ex4.getClass().getName() + " unmarshalling return: ", ex4);
                throw new UnmarshalException("error unmarshalling return", ex4);
            }
            finally {
                try {
                    streamRemoteCall.done();
                }
                catch (IOException ex5) {
                    b = false;
                }
            }
        }
        catch (RuntimeException ex6) {
            if (streamRemoteCall == null || ((StreamRemoteCall)streamRemoteCall).getServerException() != ex6) {
                b = false;
            }
            throw ex6;
        }
        catch (RemoteException ex7) {
            b = false;
            throw ex7;
        }
        catch (Error error) {
            b = false;
            throw error;
        }
        finally {
            if (n2 == 0) {
                if (UnicastRef.clientRefLog.isLoggable(Log.BRIEF)) {
                    UnicastRef.clientRefLog.log(Log.BRIEF, "free connection (reuse = " + (boolean)b + ")");
                }
                this.ref.getChannel().free(connection, (boolean)b);
            }
        }
    }
    
    protected void marshalCustomCallData(ObjectOutput objectOutput) throws IOException {
    }
    
    protected static void marshalValue(Class<?> obj, Object o, ObjectOutput objectOutput) throws IOException {
        if (obj.isPrimitive()) {
            if (obj == Integer.TYPE) {
                objectOutput.writeInt((int)o);
            }
            else if (obj == Boolean.TYPE) {
                objectOutput.writeBoolean((boolean)o);
            }
            else if (obj == Byte.TYPE) {
                objectOutput.writeByte((int)(byte)o);
            }
            else if (obj == Character.TYPE) {
                objectOutput.writeChar((int)(char)o);
            }
            else if (obj == Short.TYPE) {
                objectOutput.writeShort((int)(short)o);
            }
            else if (obj == Long.TYPE) {
                objectOutput.writeLong((long)o);
            }
            else if (obj == Float.TYPE) {
                objectOutput.writeFloat((float)o);
            }
            else if (obj == Double.TYPE) {
                objectOutput.writeDouble((double)o);
            }
            else {
                throw new Error("Unrecognized primitive type: " + obj);
            }
        }
        else {
            objectOutput.writeObject(o);
        }
    }
    
    protected static Object unmarshalValue(Class<?> obj, ObjectInput objectInput) throws IOException, ClassNotFoundException {
        if (obj.isPrimitive()) {
            if (obj == Integer.TYPE) {
                return Integer.valueOf(objectInput.readInt());
            }
            if (obj == Boolean.TYPE) {
                return Boolean.valueOf(objectInput.readBoolean());
            }
            if (obj == Byte.TYPE) {
                return Byte.valueOf(objectInput.readByte());
            }
            if (obj == Character.TYPE) {
                return Character.valueOf(objectInput.readChar());
            }
            if (obj == Short.TYPE) {
                return Short.valueOf(objectInput.readShort());
            }
            if (obj == Long.TYPE) {
                return Long.valueOf(objectInput.readLong());
            }
            if (obj == Float.TYPE) {
                return Float.valueOf(objectInput.readFloat());
            }
            if (obj == Double.TYPE) {
                return Double.valueOf(objectInput.readDouble());
            }
            throw new Error("Unrecognized primitive type: " + obj);
        }
        else {
            if (obj == String.class && objectInput instanceof ObjectInputStream) {
                return SharedSecrets.getJavaObjectInputStreamReadString().readString((ObjectInputStream)objectInput);
            }
            return objectInput.readObject();
        }
    }
    
    @Override
    public RemoteCall newCall(RemoteObject remoteObject, Operation[] array, int op, long hash) throws RemoteException {
        UnicastRef.clientRefLog.log(Log.BRIEF, "get connection");
        Connection connection = this.ref.getChannel().newConnection();
        try {
            UnicastRef.clientRefLog.log(Log.VERBOSE, "create call context");
            if (UnicastRef.clientCallLog.isLoggable(Log.VERBOSE)) {
                this.logClientCall(remoteObject, array[op]);
            }
            StreamRemoteCall streamRemoteCall = new StreamRemoteCall(connection, this.ref.getObjID(), op, hash);
            try {
                this.marshalCustomCallData(streamRemoteCall.getOutputStream());
            }
            catch (IOException ex) {
                throw new MarshalException("error marshaling custom call data");
            }
            return streamRemoteCall;
        }
        catch (RemoteException ex2) {
            this.ref.getChannel().free(connection, false);
            throw ex2;
        }
    }
    
    @Override
    public void invoke(RemoteCall remoteCall) throws Exception {
        try {
            UnicastRef.clientRefLog.log(Log.VERBOSE, "execute call");
            remoteCall.executeCall();
        }
        catch (RemoteException ex) {
            UnicastRef.clientRefLog.log(Log.BRIEF, "exception: ", ex);
            this.free(remoteCall, false);
            throw ex;
        }
        catch (Error error) {
            UnicastRef.clientRefLog.log(Log.BRIEF, "error: ", error);
            this.free(remoteCall, false);
            throw error;
        }
        catch (RuntimeException ex2) {
            UnicastRef.clientRefLog.log(Log.BRIEF, "exception: ", ex2);
            this.free(remoteCall, false);
            throw ex2;
        }
        catch (Exception ex3) {
            UnicastRef.clientRefLog.log(Log.BRIEF, "exception: ", ex3);
            this.free(remoteCall, true);
            throw ex3;
        }
    }
    
    private void free(RemoteCall remoteCall, boolean b) throws RemoteException {
        this.ref.getChannel().free(((StreamRemoteCall)remoteCall).getConnection(), b);
    }
    
    @Override
    public void done(RemoteCall remoteCall) throws RemoteException {
        UnicastRef.clientRefLog.log(Log.BRIEF, "free connection (reuse = true)");
        this.free(remoteCall, true);
        try {
            remoteCall.done();
        }
        catch (IOException ex) {}
    }
    
    void logClientCall(Object o, Object obj) {
        UnicastRef.clientCallLog.log(Log.VERBOSE, "outbound call: " + this.ref + " : " + o.getClass().getName() + this.ref.getObjID().toString() + ": " + obj);
    }
    
    @Override
    public String getRefClass(ObjectOutput objectOutput) {
        return "UnicastRef";
    }
    
    @Override
    public void writeExternal(ObjectOutput out) throws IOException {
        this.ref.write(out, false);
    }
    
    @Override
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        this.ref = LiveRef.read(in, false);
    }
    
    @Override
    public String remoteToString() {
        return Util.getUnqualifiedName(this.getClass()) + " [liveRef: " + this.ref + "]";
    }
    
    @Override
    public int remoteHashCode() {
        return this.ref.hashCode();
    }
    
    @Override
    public boolean remoteEquals(RemoteRef remoteRef) {
        return remoteRef instanceof UnicastRef && this.ref.remoteEquals(((UnicastRef)remoteRef).ref);
    }
    
    static {
        clientRefLog = Log.getLog("sun.rmi.client.ref", "transport", Util.logLevel);
        clientCallLog = Log.getLog("sun.rmi.client.call", "RMI", (boolean)AccessController.doPrivileged(new GetBooleanAction("sun.rmi.client.logCalls")));
    }
}
